#include <mpi.h>
#include <cstdio>
#include <cstring>
#include <cassert>
#include <ipc_debug.h>
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <vector>

void __attribute__((optimize("O0"))) attach_debugger(bool condition) {
    if (!condition) return;
    bool attached = false;

    // also write PID to a file
    std::ofstream os("/tmp/mpi_debug.pid");
    os << getpid() << std::endl;
    os.close();

    std::cout << "Waiting for debugger to be attached, PID: "
        << getpid() << std::endl;
    while (!attached) sleep(1);
}




int main(int argc, char **argv) {
    MPI_Init(&argc, &argv);

    int comm_size, rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &comm_size);

    const char *debug_rank = std::getenv("DEBUG_MPI_RANK");
    attach_debugger(debug_rank != NULL && rank == atoi(debug_rank));

    if (argc != 2) {
        fprintf(stderr, "Usage: %s <string>\nMake sure that the length of the string can be divided by the communicator size.\n", argv[0]);
    }


    char *arg = argv[1];
    size_t len = strlen(arg);

    assert(len % comm_size == 0);
    unsigned int characters_per_rank = len / comm_size;

    // Copy to new array to catch invalid memory accesses.
    std::vector<double> arr(characters_per_rank);
    const int start_index = characters_per_rank * rank;
    printf("Rank %i Copying array from  %i (count %i)\n", rank, start_index, characters_per_rank);
    for (uint64_t i = 0; i < characters_per_rank; i++) {
        arr[i] = static_cast<double>(argv[1][start_index + i]);
        printf("Rank %i idx %lu: %f\n", rank, i, arr[i]);
    }

    // Setup IPC DEBUG
    debug_ipc_init();
    debug_ipc_mpi_set_data_distribution(start_index);

    // Make actual assertion
    debug_ipc_assert_equal_mpi_double_array(arr.data(), characters_per_rank);

    MPI_Barrier(MPI_COMM_WORLD);
    MPI_Finalize();

}
